"""
    functions for loading datasets into PolarRoute.
    PolarRoute requires data as a pandas dataframe
    in the following format:

    long | lat | (time)* | value_1 | ... | value_n

    *time is optional

    Note:
        long and lat values must be in a EPSG:4326 projection
"""

import glob
import logging

from datetime import datetime

import xarray as xr
import pandas as pd
import numpy as np

from pyproj import Transformer
from pyproj import CRS

from polar_route.utils import date_range, timed_call


@timed_call
def load_amsr_folder(params, long_min, long_max, lat_min, lat_max, time_start, time_end):
    """
        Load AMSR2 data from a folder containing separate
        days of AMSR2 data in netCDF files and transform
        it into a format ingestable by PolarRoute

        Args:
            long_min (float): The minimum longitude of the data to be retrieved
            long_max (float): The maximum longitude of the data to be retrieved
            lat_min (float): The minimum latitude of the data to be retrieved
            lat_max (float): The maximum latitude of the data to be retrieved
            time_start (string): The start time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"
            time_end (string): The end time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"

            params (dict): A dictionary containing optional parameters. This
                function requires -

                params['folder'] (string): folder location of the AMSR2 dataset.
                    Files within the folder must be named in the form
                    *asi-AMSR2-s6250-<year><month><day>-v.5.4.nc*

        Returns:
            amsr_df (Dataframe): A dataframe containing AMSR2 Sea Ice Concentration
                data. The dataframe is of the format -

                lat | long | time | SIC
    """
    logging.debug("opening folder {}".format(params['folder']))

    # strip year/month/day from time_start and end_start strings.
    start_year = float(time_start.split('-')[0])
    start_month = float(time_start.split('-')[1])
    start_day = float(time_start.split('-')[2])

    end_year = float(time_end.split('-')[0])
    end_month = float(time_end.split('-')[1])
    end_day = float(time_end.split('-')[2])

    amsr_array = []
    time_array = []

    # iterate through all files in folder passed as a parameter
    for file in sorted(glob.glob(params['folder'] + '/*.nc')):
        # string year/month/day out of file name.
        # files in the folder must be named in the form 
        # asi-AMSR2-s6250-<year><month><day>-v.5.4.nc
        year = int(file.split('-')[-2][0:4])
        month = int(file.split('-')[-2][4:6])
        day = int(file.split('-')[-2][6:])

        if start_year <= year <= end_year:
            if start_month <= month <= end_month:
                if start_day <= day <= end_day:

                    amsr = xr.open_dataset(file)
                    logging.debug("found file {}".format(file))
                    amsr_array.append(amsr)

                    time = str(year) + '-' + str(month) + '-' + str(day)
                    time_array.append(time)

    amsr_concat = xr.concat(amsr_array, pd.Index(time_array, name="time"))
    amsr_df = amsr_concat.to_dataframe()
    amsr_df = amsr_df.reset_index()

    if ('Hemisphere' in params.keys()) and (params['Hemisphere'] == 'North'):
        in_proj = CRS('EPSG:3411')
    elif ('Hemisphere' in params.keys()) and (params['Hemisphere'] == 'South'):
        in_proj = CRS('EPSG:3412')
    else:
        in_proj = CRS('EPSG:3412')
    out_proj = CRS('EPSG:4326')

    logging.debug("reprojecting to EPSG:4326")
    x, y = Transformer.from_crs(in_proj, out_proj, always_xy=True).transform(
        amsr_df['x'].to_numpy(), amsr_df['y'].to_numpy())

    amsr_df['lat'] = y
    amsr_df['long'] = x
    amsr_df = amsr_df[['lat', 'long', 'z', 'time']]

    amsr_df = amsr_df.rename(columns={'z': 'SIC'})

    amsr_df = amsr_df[amsr_df['long'].between(long_min, long_max)]
    amsr_df = amsr_df[amsr_df['lat'].between(lat_min, lat_max)]

    logging.debug("returned {} datapoints".format(len(amsr_df.index)))

    return amsr_df


@timed_call
def load_amsr(params, long_min, long_max, lat_min, lat_max, time_start, time_end):
    """
        Load AMSR2 data from a netCDF file and transform
        it into a format ingestable by PolarRoute

        Args:
            long_min (float): The minimum longitude of the data to be retrieved
            long_max (float): The maximum longitude of the data to be retrieved
            lat_min (float): The minimum latitude of the data to be retrieved
            lat_max (float): The maximum latitude of the data to be retrieved
            time_start (string): The start time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"
            time_end (string): The end time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"

            params (dict): A dictionary containing optional parameters. This
                function requires -

                params['file'] (string): file location of the AMSR2 dataset

        Returns:
            amsr_df (Dataframe): A dataframe containing AMSR2 Sea Ice Concentration
                data. The dataframe is of the format -

                lat | long | time | SIC
    """
    logging.debug("opening file {}".format(params['file']))
    amsr = xr.open_dataset(params['file'])
    amsr = amsr.sel(time=slice(time_start, time_end))
    amsr_df = amsr.to_dataframe()
    amsr_df = amsr_df.reset_index()

    # AMSR data is in a EPSG:3412 projection and must be reprojected into
    # EPSG:4326

    logging.debug("reprojecting to EPSG:4326")
    in_proj = CRS('EPSG:3412')
    out_proj = CRS('EPSG:4326')

    x, y = Transformer.from_crs(in_proj, out_proj, always_xy=True).transform(
        amsr_df['x'].to_numpy(), amsr_df['y'].to_numpy())

    amsr_df['lat'] = y
    amsr_df['long'] = x
    amsr_df = amsr_df[['lat', 'long', 'z', 'time']]

    amsr_df = amsr_df.rename(columns={'z': 'SIC'})

    amsr_df = amsr_df[amsr_df['long'].between(long_min, long_max)]
    amsr_df = amsr_df[amsr_df['lat'].between(lat_min, lat_max)]

    logging.debug("returned {} datapoints".format(len(amsr_df.index)))
    return amsr_df


@timed_call
def load_density(params, long_min, long_max, lat_min, lat_max, time_start, time_end):
    """
        Create ice density dataframe for given time and region and put it into a format ingestable by the pyRoutePlanner.
        Data taken from Table 3 in: doi:10.1029/2007JC004254

        Args:
            long_min (float): The minimum longitude of the data to be retrieved
            long_max (float): The maximum longitude of the data to be retrieved
            lat_min (float): The minimum latitude of the data to be retrieved
            lat_max (float): The maximum latitude of the data to be retrieved
            time_start (string): The start time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"
            time_end (string): The end time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"

        Returns:
            density_df (Dataframe): A dataframe containing ice density
                data. The dataframe is of the format -

                lat | long | time | density
    """

    seasons = {1: 'su', 2: 'su', 3: 'a', 4: 'a', 5: 'a', 6: 'w', 7: 'w', 8: 'w', 9: 'sp', 10: 'sp', 11: 'sp',
               12: 'su'}
    densities = {'su': 875.0, 'sp': 900.0, 'a': 900.0, 'w': 920.0}

    def ice_density(d):
        month = d.month
        season = seasons[month]
        den = densities[season]
        return den

    start_date = datetime.strptime(time_start, "%Y-%m-%d").date()
    end_date = datetime.strptime(time_end, "%Y-%m-%d").date()

    lats = [lat for lat in np.arange(lat_min, lat_max, 0.05)]
    lons = [lon for lon in np.arange(long_min, long_max, 0.05)]
    dates = [single_date for single_date in date_range(start_date, end_date)]

    density_data = xr.DataArray(
        data=[[[ice_density(dt)
                for _ in lons]
               for _ in lats]
              for dt in dates],
        coords=dict(
            lat=lats,
            long=lons,
            time=[dt.strftime("%Y-%m-%d") for dt in dates],
        ),
        dims=("time", "lat", "long"),
        name="density",
    )

    density_df = density_data.\
        to_dataframe().\
        reset_index().\
        set_index(['lat', 'long', 'time']).reset_index()

    logging.debug("returned {} datapoints".format(len(density_df.index)))
    return density_df


@timed_call
def load_thickness(params, long_min, long_max, lat_min, lat_max, time_start, time_end):
    """
        Create ice thickness dataframe for given time and region and put it into a format ingestable by PolarRoute.
        Data taken from Table 3 in: doi:10.1029/2007JC004254

        Args:
            long_min (float): The minimum longitude of the data to be retrieved
            long_max (float): The maximum longitude of the data to be retrieved
            lat_min (float): The minimum latitude of the data to be retrieved
            lat_max (float): The maximum latitude of the data to be retrieved
            time_start (string): The start time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"
            time_end (string): The end time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"

        Returns:
            thickness_df (Dataframe): A dataframe containing ice thickness
                data. The dataframe is of the format -

                lat | long | time | thickness
    """
    thicknesses = {'Ross': {'w': 0.72, 'sp': 0.67, 'su': 1.32, 'a': 0.82, 'y': 1.07},
                   'Bellinghausen': {'w': 0.65, 'sp': 0.79, 'su': 2.14, 'a': 0.79, 'y': 0.90},
                   'Weddell E': {'w': 0.54, 'sp': 0.89, 'su': 0.87, 'a': 0.44, 'y': 0.73},
                   'Weddell W': {'w': 1.33, 'sp': 1.33, 'su': 1.20, 'a': 1.38, 'y': 1.33},
                   'Indian': {'w': 0.59, 'sp': 0.78, 'su': 1.05, 'a': 0.45, 'y': 0.68},
                   'West Pacific': {'w': 0.72, 'sp': 0.68, 'su': 1.17, 'a': 0.75, 'y': 0.79},
                   'None': {'w': 0.72, 'sp': 0.67, 'su': 1.32, 'a': 0.82, 'y': 1.07}}
    seasons = {1: 'su', 2: 'su', 3: 'a', 4: 'a', 5: 'a', 6: 'w', 7: 'w', 8: 'w', 9: 'sp', 10: 'sp', 11: 'sp',
               12: 'su'}

    def ice_thickness(d, long):
        """
            Returns ice thickness. Data taken from Table 3 in: doi:10.1029/2007JC004254
        """
        # The table has missing data points for Bellinghausen Autumn and Weddell W Winter, may require further thought
        month = d.month
        season = seasons[month]
        sea = None

        if -130 <= long < -60:
            sea = 'Bellinghausen'
        elif -60 <= long < -45:
            sea = 'Weddell W'
        elif -45 <= long < 20:
            sea = 'Weddell E'
        elif 20 <= long < 90:
            sea = 'Indian'
        elif 90 <= long < 160:
            sea = 'West Pacific'
        elif (160 <= long < 180) or (-180 <= long < -130):
            sea = 'Ross'
        else:
            sea = 'None'

        return thicknesses[sea][season]

    start_date = datetime.strptime(time_start, "%Y-%m-%d").date()
    end_date = datetime.strptime(time_end, "%Y-%m-%d").date()

    lats = [lat for lat in np.arange(lat_min, lat_max, 0.05)]
    lons = [lon for lon in np.arange(long_min, long_max, 0.05)]
    dates = [single_date for single_date in date_range(start_date, end_date)]

    thick_data = xr.DataArray(
        data=[[[ice_thickness(dt, lng)
                for lng in lons]
               for _ in lats]
              for dt in dates],
        coords=dict(
            lat=lats,
            long=lons,
            time=[dt.strftime("%Y-%m-%d") for dt in dates],
        ),
        dims=("time", "lat", "long"),
        name="thickness",
    )

    thick_df = thick_data.\
        to_dataframe().\
        reset_index().\
        set_index(['lat', 'long', 'time']).reset_index()

    logging.debug("returning {} datapoints".format(len(thick_df.index)))
    return thick_df


@timed_call
def load_baltic_thickness_density(params, long_min, long_max, lat_min, lat_max, time_start, time_end):
    """
        Create ice thickness and density dataframe for baltic route

        Args:
            long_min (float): The minimum longitude of the data to be retrieved
            long_max (float): The maximum longitude of the data to be retrieved
            lat_min (float): The minimum latitude of the data to be retrieved
            lat_max (float): The maximum latitude of the data to be retrieved
            time_start (string): The start time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"
            time_end (string): The end time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"

        Returns:
            thickness_df (Dataframe): A dataframe containing ice thickness and density
                data. The dataframe is of the format -

                lat | long | time | thickness | density
    """

    baltic_thick_data = []
    start_date = datetime.strptime(time_start, "%Y-%m-%d").date()
    end_date = datetime.strptime(time_end, "%Y-%m-%d").date()

    for single_date in date_range(start_date, end_date):
        dt = single_date.strftime("%Y-%m-%d")
        for lat in np.arange(lat_min, lat_max, 0.1):
            for lng in np.arange(long_min, long_max, 0.1):
                baltic_thick_data.append({'time': dt, 'lat': lat, 'long': lng, 'thickness': 0.3, 'density': 900.})

    baltic_thick_df = pd.DataFrame(baltic_thick_data).set_index(['lat', 'long', 'time'])
    baltic_thick_df = baltic_thick_df.reset_index()

    logging.debug("returned {} datapoints".format(len(baltic_thick_df.index)))
    return baltic_thick_df


@timed_call
def load_bsose(params, long_min, long_max, lat_min, lat_max, time_start, time_end):
    """
        Load BSOSE data from a netCDF file and transform it
        into a format ingestable by PolarRoute

        Args:
            long_min (float): The minimum longitude of the data to be retrieved
            long_max (float): The maximum longitude of the data to be retrieved
            lat_min (float): The minimum latitude of the data to be retrieved
            lat_max (float): The maximum latitude of the data to be retrieved
            time_start (string): The start time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"
            time_end (string): The end time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"

            params (dict): A dictionary containing optional parameters. This
                function requires -

                params['file'] (string): file location of the BSOSE dataset

                params['units'] (string)(optional): The units SIC will be measured
                    in. <percentage> | <fraction>. Default is percentage

        Returns:
            bsose_df (Dataframe): A dataframe containing BSOSE Sea Ice Concentration
                data. The dataframe is of the format -

                lat | long | time | SIC
    """
    logging.debug("opening file {} ".format(params['file']))
    bsose = xr.open_dataset(params['file'])
    bsose = bsose.sel(time=slice(time_start, time_end))
    bsose_df = bsose.to_dataframe()
    bsose_df = bsose_df.reset_index()

    # BSOSE data is indexed between 0:360 degrees in longitude where as the route planner
    # requires data index between -180:180 degrees in longitude
    bsose_df['long'] = bsose_df['XC'].apply(lambda x: x - 360 if x > 180 else x)
    bsose_df['lat'] = bsose_df['YC']
    bsose_df = bsose_df[['lat', 'long', 'SIarea', 'time']]

    bsose_df = bsose_df.rename(columns={'SIarea': 'SIC'})

    if 'units' in params.keys():
        if params['units'] == "percentage":
            bsose_df['SIC'] = bsose_df['SIC']*100
        if params['units'] == 'fraction':
            """""" #BSOSE source data is fractional, no transformation required
    else:
        bsose_df['SIC'] = bsose_df['SIC']*100

    bsose_df = bsose_df[bsose_df['long'].between(long_min, long_max)]
    bsose_df = bsose_df[bsose_df['lat'].between(lat_min, lat_max)]

    logging.debug("returned {} datapoints".format(len(bsose_df.index)))
    return bsose_df


@timed_call
def load_baltic_sea_ice(params, long_min, long_max, lat_min, lat_max, time_start, time_end):
    """
        Load Sea ice data for the baltic sea from a netCDF file and transform it
        into a format ingestable by PolarRoute
        Data Source: https://doi.org/10.48670/moi-00131

        Args:
            long_min (float): The minimum longitude of the data to be retrieved
            long_max (float): The maximum longitude of the data to be retrieved
            lat_min (float): The minimum latitude of the data to be retrieved
            lat_max (float): The maximum latitude of the data to be retrieved
            time_start (string): The start time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"
            time_end (string): The end time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"

            params (dict): A dictionary containing optional parameters. This
                function requires -

                params['file'] (string): file location of the Baltic ice dataset

        Returns:
            baltic_df (Dataframe): A dataframe containing Baltic Sea Ice Concentration
                data. The dataframe is of the format -

                lat | long | time | SIC
    """
    logging.debug("opening file {}".format(params['file']))
    baltic = xr.open_dataset(params['file'])
    baltic = baltic.sel(time=slice(time_start, time_end))
    baltic_df = baltic.to_dataframe()
    baltic_df = baltic_df.reset_index()

    # The Baltic ice data has spatial variables: 'lon' and 'lat'
    baltic_df['long'] = baltic_df['lon']

    baltic_df = baltic_df[['lat', 'long', 'ice_concentration', 'time']]

    baltic_df = baltic_df.rename(columns={'ice_concentration': 'SIC'})

    baltic_df = baltic_df[baltic_df['long'].between(long_min, long_max)]
    baltic_df = baltic_df[baltic_df['lat'].between(lat_min, lat_max)]

    logging.debug("returning {} datapoints".format(len(baltic_df.index)))
    return baltic_df


@timed_call
def load_bsose_depth(params, long_min, long_max, lat_min, lat_max, time_start, time_end):
    """
        Load BSOSE data from a netCDF file and transform it
        into a format ingestable by PolarRoute

        Args:
            long_min (float): The minimum longitude of the data to be retrieved
            long_max (float): The maximum longitude of the data to be retrieved
            lat_min (float): The minimum latitude of the data to be retrieved
            lat_max (float): The maximum latitude of the data to be retrieved
            time_start (string): The start time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"
            time_end (string): The end time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"

            params (dict): A dictionary containing optional parameters. This
                function requires -

                params['file'] (string): file location of the BSOSE dataset

        Returns:
            bsose_df (Dataframe): A dataframe containing BSOSE bathymetry
                data. The dataframe is of the format -

                lat | long | time | elevation
    """
    logging.debug("opening file {} ".format(params['file']))
    bsose = xr.open_dataset(params['file'])
    bsose = bsose.sel(time=slice(time_start, time_end))
    bsose_df = bsose.to_dataframe()
    bsose_df = bsose_df.reset_index()

    # BSOSE data is indexed between 0:360 degrees in longitude where as the route planner
    # requires data index between -180:180 degrees in longitude
    bsose_df['long'] = bsose_df['XC'].apply(lambda x: x - 360 if x > 180 else x)
    bsose_df['lat'] = bsose_df['YC']
    bsose_df = bsose_df[['lat', 'long', 'Depth', 'time']]
    bsose_df['Depth'] = -bsose_df['Depth']

    bsose_df = bsose_df.rename(columns={'Depth': 'elevation'})

    bsose_df = bsose_df[bsose_df['long'].between(long_min, long_max)]
    bsose_df = bsose_df[bsose_df['lat'].between(lat_min, lat_max)]

    logging.debug("returning {} datapoints".format(len(bsose_df.index)))
    return bsose_df


@timed_call
def load_gebco(params, long_min, long_max, lat_min, lat_max, time_start, time_end):
    """
        Load GEBCO data from a netCDF file and transform it
        into a format ingestable by the pyRoutePlanner

        Args:
            long_min (float): The minimum longitude of the data to be retrieved
            long_max (float): The maximum longitude of the data to be retrieved
            lat_min (float): The minimum latitude of the data to be retrieved
            lat_max (float): The maximum latitude of the data to be retrieved
            time_start (string): The start time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"
            time_end (string): The end time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"

            params (dict): A dictionary containing optional parameters. This
                function requires -

                params['file'] (string): file location of the GEBCO dataset
                params['downsample_factors'] ([int,int]): Downsample factors in horizontal and vertical respectively.

        Returns:
            gebco_df (Dataframe): A dataframe containing GEBCO elevation
                data. The dataframe is of the format -

                lat | long | time | elevation 
    """
    logging.debug("opening file {}".format(params['file']))
    gebco = xr.open_dataset(params['file'])

    
    if 'downsample_factors' in params.keys():
        logging.debug("downsampling dataset by a factor of [{},{}]".format(params['downsample_factors'][0],
            params['downsample_factors'][1] ))
        elev = gebco['elevation'][::params['downsample_factors'][0],::params['downsample_factors'][1]]
        gebco = xr.Dataset()
        gebco['elevation'] = elev

    gebco_df = gebco.to_dataframe()
    gebco_df = gebco_df.reset_index()
    gebco_df = gebco_df.rename(columns={'lon': 'long'})

    gebco_df = gebco_df[gebco_df['long'].between(long_min, long_max)]
    gebco_df = gebco_df[gebco_df['lat'].between(lat_min, lat_max)]

    logging.debug("returned {} datapoints".format(len(gebco_df.index)))
    return gebco_df


@timed_call
def load_sose_currents(params, long_min, long_max, lat_min, lat_max, time_start, time_end):
    """
        Load SOSE current data from a netCDF file and#
        transform it into a format that is ingestable
        by the pyRoutePlanner

        Args:
            long_min (float): The minimum longitude of the data to be retrieved
            long_max (float): The maximum longitude of the data to be retrieved
            lat_min (float): The minimum latitude of the data to be retrieved
            lat_max (float): The maximum latitude of the data to be retrieved
            time_start (string): The start time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"
            time_end (string): The end time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"

            params (dict): A dictionary containing optional parameters. This
                function requires -

                params['file'] (string): file location of the SOSE dataset

                params['units'] (string)(optional) : The units of measurements
                    uC and vC will be given in - <km/h> | <m/s>. Default is m/s

        Returns:
            sose_df (Dataframe): A dataframe containing SOSE current
                data. The dataframe is of the format -

                lat | long | time | uC | vC
    """
    logging.debug("opening file {}".format(params['file']))
    sose = xr.open_dataset(params['file'])
    sose_df = sose.to_dataframe()
    sose_df = sose_df.reset_index()

    # SOSE data is indexed between 0:360 degrees in longitude where as the route planner
    # requires data index between -180:180 degrees in longitude
    sose_df['long'] = sose_df['lon'].apply(lambda x: x - 360 if x > 180 else x)
    sose_df = sose_df[['lat', 'long', 'uC', 'vC']]

    sose_df = sose_df[sose_df['long'].between(long_min, long_max)]
    sose_df = sose_df[sose_df['lat'].between(lat_min, lat_max)]

    if 'units' in params.keys():
        if params['units'] == "km/h":
            sose_df['uC'] = sose_df['uC'] * 3.6
            sose_df['vC'] = sose_df['vC'] * 3.6
        if params['units'] == 'm/s':
            """""" # SOSE source data is in m/s, no transformation required

    logging.debug("returning {} datapoints".format(len(sose_df.index)))
    return sose_df


@timed_call
def load_baltic_currents(params, long_min, long_max, lat_min, lat_max, time_start, time_end):
    """
        Load Baltic Sea current data from a netCDF file and
        transform it into a format that is ingestable
        by the pyRoutePlanner
        Data source: https://doi.org/10.48670/moi-00013

        Args:
            long_min (float): The minimum longitude of the data to be retrieved
            long_max (float): The maximum longitude of the data to be retrieved
            lat_min (float): The minimum latitude of the data to be retrieved
            lat_max (float): The maximum latitude of the data to be retrieved
            time_start (string): The start time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"
            time_end (string): The end time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"

            params (dict): A dictionary containing optional parameters. This
                function requires -

                params['file'] (string): file location of the Baltic current dataset

        Returns:
            bc_df (Dataframe): A dataframe containing Baltic Sea current
                data. The dataframe is of the format -

                lat | long | time | uC | vC
    """
    logging.debug("opening file {}".format(params['file']))
    bc = xr.open_dataset(params['file'])
    bc_df = bc.to_dataframe()
    bc_df = bc_df.reset_index()

    bc_df = bc_df[['latitude', 'longitude', 'uo', 'vo']]

    bc_df = bc_df.rename(columns={'longitude': 'long', 'latitude': 'lat', 'uo': 'uC', 'vo': 'vC'})

    bc_df = bc_df[bc_df['long'].between(long_min, long_max)]
    bc_df = bc_df[bc_df['lat'].between(lat_min, lat_max)]

    logging.debug("returned {} datapoints".format(len(bc_df.index)))
    return bc_df


@timed_call
def load_modis(params, long_min, long_max, lat_min, lat_max, time_start, time_end):
    """
        Load MODIS data from a netCDF file and transform it
        into a format that is ingestable by the pyRoutePlanner

        Args:
            long_min (float): The minimum longitude of the data to be retrieved
            long_max (float): The maximum longitude of the data to be retrieved
            lat_min (float): The minimum latitude of the data to be retrieved
            lat_max (float): The maximum latitude of the data to be retrieved
            time_start (string): The start time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"
            time_end (string): The end time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"

            params (dict): A dictionary containing optional parameters. This
                function requires -
  
                params['file'] (string): file location of the MODIS dataset

        Returns:
            modis_df (Dataframe): A dataframe containing MODIS Sea Ice Concentration
                data. The dataframe is of the format -

                lat | long | time | SIC | cloud
    """
    logging.debug("opening file {}".format(params['file']))
    modis = xr.open_dataset(params['file'])
    modis_df = modis.to_dataframe()
    modis_df = modis_df.reset_index()

    # MODIS Sea Ice Concentration data is partially obscured by cloud cover.
    # Where a data point indicates that there is cloud cover above it,
    # set the SIC of that datapoint to NaN
    modis_df['iceArea'] = np.where(modis_df['cloud'] == 1, np.NaN, modis_df['iceArea'])
    modis_df = modis_df.rename(columns={'iceArea': 'SIC'})
    modis_df['SIC'] = modis_df['SIC']*10.

    modis_df = modis_df[modis_df['long'].between(long_min, long_max)]
    modis_df = modis_df[modis_df['lat'].between(lat_min, lat_max)]

    logging.debug("returned {} datapoints".format(len(modis_df.index)))
    return modis_df


@timed_call
def load_era5_wind(params, long_min, long_max, lat_min, lat_max, time_start, time_end):
    """
        Load era5 wind data from a netCDF file and transform it
        into a format that is ingestable by the pyRoutePlanner

        Args:
            long_min (float): The minimum longitude of the data to be retrieved
            long_max (float): The maximum longitude of the data to be retrieved
            lat_min (float): The minimum latitude of the data to be retrieved
            lat_max (float): The maximum latitude of the data to be retrieved
            time_start (string): The start time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"
            time_end (string): The end time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"

            params (dict): A dictionary containing optional parameters. This
                function requires -

                params['file'] (string): file location of the era5 dataset

        Returns:
            era5_wind_df (Dataframe): A dataframe containing era5 wind
                data. The dataframe is of the format -

                lat | long | time | u10 | v10
    """
    logging.debug("opening file {}".format(params['file']))
    era5_wind = xr.open_dataset(params['file'])

    # era5_wind data is available in monthly slices, not daily. 
    # time_start is set to the start of the given month to ensure that data is loaded.
    time_start_split = time_start.split('-')
    time_start = time_start_split[0] + "-" + time_start_split[1] + "-01"

    era5_wind = era5_wind.sel(time=slice(time_start, time_end))

    era5_wind_df = era5_wind.to_dataframe()
    era5_wind_df = era5_wind_df.reset_index()

    era5_wind_df = era5_wind_df.rename(columns={'longitude': 'long', 'latitude': 'lat'})
    era5_wind_df = era5_wind_df[era5_wind_df['long'].between(long_min, long_max)]
    era5_wind_df = era5_wind_df[era5_wind_df['lat'].between(lat_min, lat_max)]

    logging.debug("returned {} datapoints".format(len(era5_wind.index)))
    return era5_wind_df


@timed_call
def load_north_sea_currents(params, long_min, long_max, lat_min, lat_max, time_start, time_end):
    """
        Args:
            long_min (float): The minimum longitude of the data to be retrieved
            long_max (float): The maximum longitude of the data to be retrieved
            lat_min (float): The minimum latitude of the data to be retrieved
            lat_max (float): The maximum latitude of the data to be retrieved
            time_start (string): The start time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"
            time_end (string): The end time of the data to be retrieved,
                must be given in the format "YYYY-MM-DD"

            params (dict): A dictionary containing optional parameters. This
                function requires -

                params['file'] (string): file location of the Baltic current dataset

        Returns:
            bc_df (Dataframe): A dataframe containing Baltic Sea current
                data. The dataframe is of the format -

                lat | long | time | uC | vC
    """
    logging.debug("opening file {}".format(params['file']))
    bc = xr.open_dataset(params['file'])
    bc_df = bc.to_dataframe()

    bc_df = bc_df.reset_index()
    bc_df = bc_df[['lat', 'lon', 'U', 'V']]
    bc_df = bc_df.rename(columns={'lon': 'long', 'lat': 'lat', 'U': 'uC', 'V': 'vC'})
    logging.debug("returned {} datapoints".format(len(bc_df.index)))
    return bc_df


@timed_call
def load_oras5(params, long_min, long_max, lat_min, lat_max, time_start, time_end):
    logging.debug("opening file {}".format(params['file']))
    oras5 = xr.open_dataset(params['file'])
    oras5 = oras5.sel(time=slice(time_start, time_end))
    oras5_df = oras5.to_dataframe()
    oras5_df = oras5_df.reset_index()

    oras5_df = oras5_df.rename(columns={'uo': 'uC','vo': 'vC','longitude':'long','latitude':'lat'})
    oras5_df = oras5_df[oras5_df['long'].between(long_min, long_max)]
    oras5_df = oras5_df[oras5_df['lat'].between(lat_min, lat_max)]
    logging.debug("returned {} datapoints".format(len(oras5_df.index)))
    return oras5_df



"""
Outlined in this section we will discuss the usage of the CellBox functionality
of the PolarRoute package. In this series of class distributions we house our discrete
representation of input data. In each CellBox we determine the mean and variance of
the information governing our numerical world, this includes and is not limited to:
Ocean Currents, Sea Ice Concentration and Bathymetric depth.

Example:
    An example of running this code can be executed by running the following
    in a ipython/Jupyter Notebook::

        from polar_route import cellbox
        ....

Note:
    CellBoxes are intended to be constructed by and used within a Mesh
    object. The methods provided are to extract information for CellBoxes
    contained within a Mesh.
"""

from shapely.geometry import Polygon
import numpy as np
import pandas as pd


class CellBox:
    """
    A CellBox is a collection of data-points contained within a given geo-spatial/temporal
    boundary. Information about any given value of a CellBox is calculated from 
    the mean of all data points of that type within those bounds. CellBoxes may
    be split into smaller CellBoxes and the data points within distributed
    between the newly created CellBoxes so as to construct a non-uniform mesh
    of CellBoxes, such as within a Mesh.


    Attributes:
        lat (float): The latitude of the top-left corner of the CellBox
        long (float): The longitude of the top-left corner of the CellBox
        width (float): The width of the CellBox, given in degrees longitude
        height (float): The height of the CellBox, given in degrees latitude

    Note:
        All geospatial boundaries of a CellBox are given in a 'EPSG:4326' projection
    """
    split_depth = 0

    def __init__(self, lat, long, width, height, splitting_conditions=[], j_grid=False):
        """

            Args:
                lat (float): The latitude of the top-left corner of the CellBox
                long (float): The longitude of the top-left corner of the CellBox
                width (float): The width of the CellBox, given in degrees longitude
                height (float): The height of the CellBox, given in degrees latitude
                splitting_conditions (list<(dict)>): conditions in which the CellBox
                    will be split into 4 smaller CellBoxes.

                    splitting_conditions are of the form -
                        {
                            <value>:{
                                "threshold" (float):,\n
                                "upperBound" (float):,\n
                                "lowerBound" (float)
                            }
                        }

                j_grid (bool): True if the CellBox should be constructed using the
                    format of the original Java codebase
        """
        # Box information relative to bottom left
        self.lat = lat
        self.long = long
        self.width = width
        self.height = height

        self._data_points = pd.DataFrame()

        self.minimum_datapoints = 10
        self._splitting_conditions = splitting_conditions
        self._value_fill_types = dict()

        self._value_out_types = dict()

        # For initial implementation of land based from Java codebase.
        self._j_grid = j_grid
        self.land_locked = False
        self.grid_uc = 0
        self.grid_vc = 0
        self.x_coord = 0
        self.y_coord = 0
        self.focus = ""
        self.min_depth = 10
        self._current_points = pd.DataFrame()

    # Functions used for getting data from a cellBox
    def getcx(self):
        """
            returns x-position of the centroid of the cellbox

            Returns:
                cx (float): the x-position of the top-left corner of the CellBox
                    given in degrees longitude.
        """
        return self.long + self.width/2

    def getcy(self):
        """
            returns y-position of the centroid of the cellbox

            Returns:
                cy (float): the y-position of the top-left corner of the CellBox
                    given in degrees latitude.
        """
        return self.lat + self.height/2

    def getdcx(self):
        """
            returns x-distance from the edge to the centroid of the cellbox

            Returns:
                dcx (float): the x-distance from the edge of the CellBox to the 
                    centroid of the CellBox. Given in degrees longitude
        """
        return self.width/2

    def getdcy(self):
        """
            returns y-distance from the edge to the centroid of the cellbox

            Returns:
                dxy (float): the y-distance from the edge of the CellBox to the
                    centroid of the CellBox. Given in degrees latitude
        """
        return self.height/2

    def get_data_names(self):
        """
            Returns the data names of all values which have been added to this CellBox

            Returns:
                data_names (list<(String)>): A list of all the names of data types which
                    have been added to this CellBox
        """
        data_names = list(self._data_points.columns)

        to_remove = ['lat', 'long', 'time']
        for item in to_remove:
            data_names.remove(item)

        return data_names

    # TODO: getter / setter
    def get_data_points(self, values=[]):
        """
            Returns a dataframe of containing values specified in parameter 'values'.
            If values is empty, return a dataframe containing all datapoints within
            the CellBox.

            Args:
                values (list<string>): datapoints within the CellBox to be included
                    in the returned dataframe

            Returns:
                data_points (Dataframe): a dataframe of datapoints within the CellBox.
                The dataframe is of the form -

                    long | lat | time | value_1 | ... | value_n
        """
        if len(values) == 0:
            return self._data_points
        else:
            data_points = pd.DataFrame()

            # TODO: review, this looks rather inefficient through recursions
            for value in values:
                data_points = pd.concat(
                    [data_points, self.get_data_points().dropna(subset=[value])], axis=0)

            columns = ['lat', 'long', 'time'] + values
            return data_points[columns]

    # TODO: getter / setter
    # TODO: caller can aggregate, this looks like an anti-pattern
    def get_value(self, value_name, value_type="MEAN"):
        """
            returns the mean value of the datapoints within this cellbox
            specified by the parameter 'value'.

            Args:
                value_name (string): The value type requested

                value_type (string): The output type of a value requested.
                    value_type may be < MEAN | MIN | MAX >. If none is given
                    a default of MIN is used.

            Returns:
                value (float): The mean of all data_points of type 'value'
                    within this CellBox
        """
        data_frame = self.get_data_points(values=[value_name])

        if value_type == "MIN":
            value = data_frame[value_name].min()
        elif value_type == "MAX":
            value = data_frame[value_name].max()
        else:  # value_type == MEAN
            value = data_frame[value_name].mean()

        return value

    def get_bounds(self):
        """
            returns the bounds of this cellbox

            Returns:
                bounds (list<tuples>): The geo-spatial boundaries of this CellBox.
        """
        bounds = [[self.long, self.lat],
                    [self.long, self.lat + self.height],
                    [self.long + self.width, self.lat + self.height],
                    [self.long + self.width, self.lat],
                    [self.long, self.lat]]
        return bounds

    def get_value_out_types(self):
        """
            TODO
        """
        return dict(self._value_out_types)

    def get_value_fill_types(self):
        """
            TODO
        """
        return dict(self._value_fill_types)

    def set_minimum_datapoints(self, minimum_datapoints):
        """
            TODO
        """
        self.minimum_datapoints = minimum_datapoints

    # Functions used for adding data to a cellbox
    def add_data_points(self, new_data_points):
        """
            adds a dataframe containing datapoints to the dataframe
            of datapoints within this CellBox

            Args:
                new_data_points (Dataframe): A dataframe of data_points to be added
                to this CellBox. new_data_points must be of the format -

                    lat | long | (time)* | value_1 | ... | value_n
        """
        self._data_points = pd.concat([self._data_points, new_data_points], axis=0)

    def add_splitting_condition(self, splitting_condition):
        """
            adds a dictionary containing a splitting condition to the
            list of splitting conditions contained within this CellBox

            Args:
                splitting_condition (dict): a splitting condition to be added to
                this CellBox. splitting_condition must be of the form -

                splitting condition is of form:
                {<value>: {
                    'threshold': (float) ...,
                    'upperbound': (float) ...,
                    'lowerbound' (float) ...:
                }}
        """
        self._splitting_conditions = self._splitting_conditions + [splitting_condition]

    def add_value_output_type(self, value_out_type):
        """
            appends a dictionary mapping values in a cellbox to there output types to the cellboxes
            internal memory of values to output type mappings.

            Args:
                value_out_type (string): A dictionary containing a mapping of a value held within
                the cellbox to its output type. An output type may be either MEAN, MIN or MAX.
                If no output type is defined for a value in a cellbox, this defaults as MEAN.

                {
                    <value>: < MEAN | MIN | MAX >,
                    ...
                }
        """
        self._value_out_types.update(value_out_type)

    def set_value_fill_types(self, value_fill_types):
        """
            TODO
        """
        self._value_fill_types = value_fill_types


    # Functions used for splitting a cellbox
    def value_should_be_split(self, value, threshold, lowerbound, upperbound):
        """
            returns true or false depending on whether a splitting condition associated
            with parameter 'value' should cause the cellbox to be split according to the
            parameters 'threshold', 'upperbound' and 'lowerbound'.

            Args:
                value (string): the name of a value a splitting condition is checked against.
                threshold (float): The threshold at which data_points of type 'value' within
                    this CellBox are checked to be either above or below.
                lowerbound (float): The lowerbound of acceptable percentage of data_points of
                    type value within this CellBox that are above 'threshold'.
                upperbound (float): the upperbound of acceptable percentage of data_points of
                    type value within this CellBox that are above 'threshold'.

            Returns:
                should_be_split (bool): True if the splitting_condition given would result in
                    this CellBox being split.
        """
        data_limit = self.minimum_datapoints

        data_points = self.get_data_points(values=[value])

        if data_points.shape[0] < data_limit:
            return False

        prop_over = data_points.loc[data_points[value] > threshold]

        proportion_over_x_percent = prop_over.shape[0] / data_points.shape[0]
        return lowerbound < proportion_over_x_percent < upperbound

    def value_hom_condition(self, value, threshold, lowerbound, upperbound):
        """
            returns 'CLR', 'HET' or 'HOM' dependant on the distribution of
            datapoints contained within.

            Args:
                value (string): the name of a value a splitting condition is checked against.
                threshold (float): The threshold at which data points of type 'value' within
                    this CellBox are checked to be either above or below
                lowerbound (float): The lowerbound of acceptable percentage of data_points of
                    type value within this CellBox that are above 'threshold'
                upperbound (float): the upperbound of acceptable percentage of data_points of
                    type value within this CellBox that are above 'threshold'

            Returns:
                hom_condition (string): The homogeniety condtion of this CellBox by given parameters
                    hom_condition is of the form -

                CLR = the proportion of data points within this cellbox over a given
                    threshold is lower than the lowerbound
                HOM = the proportion of data points within this cellbox over a given
                    threshold is higher than the upperbound
                MIN = the cellbox contains less than a minimum number of data points

                HET = the proportion of data points within this cellbox over a given
                    threshold if between the upper and lower bound
        """
        data_limit = 4
        if self._j_grid:
            data_limit = 3000

        data_points = self.get_data_points([value])

        if data_points.shape[0] < data_limit:
            return "MIN"

        over_threshold = data_points.loc[data_points[value] > threshold]

        prop_over = over_threshold.shape[0] / data_points.shape[0]
        if prop_over <= lowerbound:
            return "CLR"
        if prop_over >= upperbound:
            return "HOM"
        return "HET"

    def hom_condition(self):
        """
            The total homogeneity condition of this CellBox, determined by
            all splitting_conditions within this CellBox.

            Returns:
                hom_condition (string): The homogeneity condition of this CellBox.
                    hom_condition is of the form -

                    CLR = the proportion of datapoints within this CellBox over a given
                        threshold is lower than the lowerbound
                    HOM = the proportion of datapoints within this cellbox over a given
                        threshold is higher than the upperbound
                    MIN = the cellbox contains less than a minimum number of datapoints

                    HET = the proportion of datapoints within this cellbox over a given
                        threshold if between the upper and lower bound

        """
        hom_conditions = []

        for splitting_condition in self._splitting_conditions:
            value = list(splitting_condition.keys())[0]
            threshold = float(splitting_condition[value]['threshold'])
            upperbound = float(splitting_condition[value]['upperBound'])
            lowerbound = float(splitting_condition[value]['lowerBound'])

            hom_conditions.append(self.value_hom_condition(value, threshold, lowerbound, upperbound))

        if "HOM" in hom_conditions:
            return "HOM"
        if "MIN" in hom_conditions:
            return "MIN"
        if "HET" in hom_conditions: 
            return "HET"
        if hom_conditions.count("CLR") == len(hom_conditions):
            return "CLR"

        return "ERR"

    def should_be_split(self):
        """
            returns true or false depending on if any of the splitting conditions
            on values contained within this cellbox dictate that the cellbox
            should be split

            DEPRECATED - use should split instead.
        """

        # if a j_grid has been generated, use a different function to determine splitting
        if self._j_grid:
            splitting_percentage = 0.12
            split_min_prop = 0.05
            split_max_prop = 0.85
            return self.should_we_split(splitting_percentage, split_min_prop, split_max_prop)

        split = False
        for splitting_condition in self._splitting_conditions:
            value = list(splitting_condition.keys())[0]
            threshold = float(splitting_condition[value]['threshold'])
            upperbound = float(splitting_condition[value]['upperBound'])
            lowerbound = float(splitting_condition[value]['lowerBound'])
            split = split or self.value_should_be_split(value, threshold, lowerbound, upperbound)
        return split

    def should_split(self):
        """
            determines if a cellbox should be split based on the homogeneity
            condition of each data type contained within. The homogeneity condition
            of values within this cellbox is calculated using the function
            'value_hom_condition()'

            if ANY data returns 'HOM':
                do not split
            if ANY data returns 'MIN':
                do not split
            if ALL data returns 'CLR':
                do not split
            else (mixture of CLR & HET):
                split

            Returns:
                should_split (bool): True if the splitting_conditions of this CellBox
                    will result in the CellBox being split.
        """
        hom_conditions = []

        for splitting_condition in self._splitting_conditions:
            value = list(splitting_condition.keys())[0]
            threshold = float(splitting_condition[value]['threshold'])
            upperbound = float(splitting_condition[value]['upperBound'])
            lowerbound = float(splitting_condition[value]['lowerBound'])

            hom_conditions.append(self.value_hom_condition(value,threshold,lowerbound,upperbound))

        if "HOM" in hom_conditions:
            return False
        if "MIN" in hom_conditions:
            return False
        if hom_conditions.count("CLR") == len(hom_conditions):
            return False

        return True

    def split(self):
        """
            splits the current cellbox into 4 corners, returns as a list of cellbox objects.

            Returns:
                split_boxes (list<CellBox>): The 4 corner cellboxes generates by splitting
                    this current cellboxes and dividing the data_points contained between.
        """

        # split_boxes = [{}, {}, {}, {}]

        half_width = self.width / 2
        half_height = self.height / 2

        # create 4 new cellBoxes
        north_west = CellBox(self.lat + half_height, self.long, half_width, half_height,
                             splitting_conditions=self._splitting_conditions, j_grid=self._j_grid)
        north_east = CellBox(self.lat + half_height, self.long + half_width, half_width, half_height,
                             splitting_conditions=self._splitting_conditions, j_grid=self._j_grid)
        south_west = CellBox(self.lat, self.long, half_width, half_height,
                             splitting_conditions=self._splitting_conditions, j_grid=self._j_grid)
        south_east = CellBox(self.lat, self.long + half_width, half_width, half_height,
                             splitting_conditions=self._splitting_conditions, j_grid=self._j_grid)

        split_boxes = [north_west, north_east, south_west, south_east]

        for split_box in split_boxes:
            split_box.split_depth = self.split_depth + 1

            # Split dataPoints per box
            long_loc = self._data_points.loc[(self._data_points['long'] > split_box.long) &
                                             (self._data_points['long'] <= (split_box.long + split_box.width))]
            lat_long_loc = long_loc.loc[(long_loc['lat'] > split_box.lat) &
                                        (long_loc['lat'] <= (split_box.lat + split_box.height))]

            split_box.add_data_points(lat_long_loc)

            split_box.add_value_output_type(self.get_value_out_types())
            split_box.set_value_fill_types(self.get_value_fill_types())
            split_box.set_minimum_datapoints(self.minimum_datapoints)

            for value in split_box.get_value_fill_types().keys():
                if value in split_box.get_value_out_types().keys():
                    value_output_type = split_box.get_value_out_types()['value']
                else:
                    value_output_type = "MEAN"

                if np.isnan(split_box.get_value(value)):
                    if split_box.get_value_fill_types()[value] == "zero":
                        fill_value = 0
                    elif split_box.get_value_fill_types()[value] == "parent":
                        fill_value = self.get_value(value, value_output_type)
                    else:
                        fill_value = np.nan
              
                    if not np.isnan(fill_value):
                        datapoint = [[split_box.getcy(), split_box.getcx(), fill_value]]
                        fill_df = pd.DataFrame(datapoint, columns=['lat','long',value])

                        split_box.add_data_points(fill_df)

            # if parent box is land, all child boxes are considered land
            if self.land_locked:
                split_box.land_locked = True

            if self._j_grid:
                split_box.grid_uc = self.grid_uc
                split_box.grid_vc = self.grid_vc

                # set gridCoord of split boxes equal to parent.
                split_box.set_grid_coord(self.x_coord, self.y_coord)

                # create focus for split boxes.
                split_box.set_focus(self.get_focus().copy())
                split_box.add_to_focus(split_boxes.index(split_box))

        return split_boxes

    # Misc

    def contains_point(self, lat, long):
        """
            Returns true if a given lat/long coordinate is contained within this cellBox.

            Args:
                lat (float): latitude of a given point
                long (float): longitude of a given point

            Returns:
                contains_points (bool): True if this CellBox contains a point given by
                    parameters (lat, long)
        """
        if (lat >= self.lat) & (lat <= self.lat + self.height):
            if (long >= self.long) & (long <= self.long + self.width):
                return True
        return False

    def __str__(self):
        '''
            Converts a cellBox to a String which may be printed to console for debugging purposes
            TODO
        '''
        cellbox_str = "TODO"
        return cellbox_str

    def to_json(self):
        '''
            convert cellBox to JSON

            The returned object is of the form -

                {
                    "geometry" (String): POLYGON(...),\n
                    "cx" (float): ...,\n
                    "cy" (float): ...,\n
                    "dcx" (float): ...,\n
                    "dcy" (float): ..., \n
                    \n
                    "value_1" (float): ...,\n
                    ...,\n
                    "value_n" (float): ...
                }
            Returns:
                cell_json (dict): A JSON parsable dictionary representation of this CellBox
        '''
        cell_json = {
            "geometry": str(Polygon(self.get_bounds())),
            'cx': float(self.getcx()),
            'cy': float(self.getcy()),
            'dcx': float(self.getdcx()),
            'dcy': float(self.getdcy())
        }

        if self._j_grid:
            cell_json['uC'] = self.grid_uc
            cell_json['vC'] = self.grid_vc
            cell_json['SIC'] = self.get_value('SIC')
            cell_json['elevation'] = 0 if self.is_land_m() else -100
            cell_json['thickness'] = 2
            cell_json['density'] = 875
        else:
            for value in self.get_data_names():
                if value in self.get_value_out_types().keys():
                    cell_json[value] = float(self.get_value(value, 
                        self.get_value_out_types()[value]))
                else:
                    cell_json[value] = float(self.get_value(value))

        return cell_json

    def contains_land(self):
        """
            DEPRECATED - Land mask are now calculated based on the average depth of a cell.

            Returns True if any icepoint within the cell has a
            depth less than the specified minimum depth.
        """

        if self._j_grid:
            return self.is_land_m()

        depth_list = self._data_points.dropna(subset=['depth'])['depth']

        if (depth_list > self.min_depth).any():
            return True
        return False

    def is_land(self):
        """
            DEPRECATED - land masked are now calculated based on the average depth of a cell

            Returns True if all icepoints within the cell have a
            depth less than the specified minimum depth.
        """
        if self._j_grid:
            return self.is_land_m()

        depth_list = self._data_points.dropna(subset=['depth'])['depth']
        if (depth_list > self.min_depth).all():
            return True
        return False

    # Functions used for j_grid regression testing.
    def set_grid_coord(self, xpos, ypos):
        """
            sets up initial grid-coordinate when creating a j_grid

            for use in j_grid regression testing
        """
        self.x_coord = xpos
        self.y_coord = ypos

    def set_focus(self, focus):
        """
            initialize the focus of this cellbox

            for use in j_grid regression testing
        """
        self.focus = focus

    def add_to_focus(self, focus):
        """
            append additional information to the focus of this cellbox
            to be used when splitting.

            for use in j_grid regression testing
        """
        self.focus.append(focus)

    def get_focus(self):
        """
            returns the focus of this cellbox

            for use in j_grid regression testing
        """
        return self.focus

    def grid_coord(self):
        """
            returns a string representation of the grid_coord of this cellbox

            for use in j_grid regression testing
        """
        return "(" + str(int(self.x_coord)) + "," + str(int(self.y_coord)) + ")"

    def node_string(self):
        """
            returns a string representing the node of this cellbox


            for use in j_grid regression testing
        """
        node_string = self.grid_coord() + " F:" + str(len(self.get_focus()))

        focus_string = "["
        for focus in self.get_focus():
            focus_string += str(focus) + " "
        focus_string += "]"
        return node_string + " " + focus_string

    def mesh_dump(self):
        """
            returns a string representing all the information stored in the mesh
            of this cellbox

            for use in j_grid regression testing
        """
        mesh_dump = ""
        mesh_dump += self.node_string() + "; "  # add node string
        mesh_dump += "0 "
        mesh_dump += str(self.getcy()) + ", " + str(self.getcx()) + "; "  # add lat,lon
        ice_area = self.get_value('SIC')
        if np.isnan(ice_area):
            ice_area = 0
        mesh_dump += str(ice_area) + "; "  # add ice area
        if np.isnan(self.grid_uc):
            mesh_dump += str(0) + ", " + str(0) + ", "
        else:
            mesh_dump += str(self.grid_uc) + ", " + str(self.grid_vc) + ", "
        mesh_dump += str(self.get_data_points(['SIC']).shape[0])
        mesh_dump += "\n"

        return mesh_dump

    def add_current_points(self, current_points):
        """
            updates the current points contained within this cellBox to a pandas
            dataframe provided by parameter currentPoints.

            Required for j_grid creation

            for use in j_grid regression testing
        """
        self._current_points = current_points.dropna()
        self.grid_uc = self._current_points['uC'].mean()
        self.grid_vc = self._current_points['vC'].mean()

        self._data_points = pd.concat([self._data_points, current_points], axis=0)

    def set_land(self):
        """
            sets attribute 'land_locked' of a cellBox based on the proportion
            of current vectors contained within it that are not empty.

            Only to be used on un-split cells

            for use in j_grid regression testing

            TODO requires changing to deem a cell land if the datapoint closest
            to the centre of the cell is nan
        """
        if self.split_depth == 0:  # Check if a cell has not been split
            total_currents = self._current_points.dropna()
            watermin = 112.5

            if total_currents.shape[0] < watermin:
                self.land_locked = True

    def is_land_m(self):
        """
            returns true/false dependent on if this cellbox is considered land

            used for j_grid regression testing
        """
        return self.land_locked

    def should_we_split(self, splitting_percentage, split_min_prop, split_max_prop):
        """
            returns true/false dependent on if this cellbox should be split

            used for j_grid regression testing
        """
        if not self._j_grid:
            return self.should_be_split()
        data_limit = 3000

        ice_points = self._data_points.dropna(subset=['iceArea'])

        if ice_points.shape[0] < data_limit:
            return False

        prop_over = ice_points.loc[ice_points['iceArea'] > splitting_percentage]

        proportion_over_x_percent = prop_over.shape[0] / ice_points.shape[0]
        return split_min_prop < proportion_over_x_percent < split_max_prop



import logging
import time
import tracemalloc

from datetime import timedelta
from functools import wraps

"""
Utilities that might be of use
"""


def date_range(start_date, end_date):
    for n in range(int((end_date - start_date).days)):
        yield start_date + timedelta(n)


def memory_trace(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        tracemalloc.start(20)
        res = func(*args, **kwargs)
        snapshot = tracemalloc.take_snapshot()
        top_stats = snapshot.statistics('traceback')

        stat = top_stats[0]
        logging.info("{} memory blocks: {.1f} KiB".
                     format(stat.count, stat.size / 1024))
        logging.info("\n".join(stat.traceback.format()))
        return res
    return wrapper


def timed_call(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        start = time.perf_counter()
        res = func(*args, **kwargs)
        end = time.perf_counter()
        logging.info("Timed call to {} took {:02f} seconds".
                     format(func.__name__, end - start))
        return res
    return wrapper


# CLI utilities
def setup_logging(func,
                  log_format="[%(asctime)-17s :%(levelname)-8s] - %(message)s"):
    """Wraps a CLI endpoint and sets up logging for it

    This is probably not the smoothest implementation, but it's an educational
    one for people who aren't aware of decorators and how they're implemented.
    In addition, it supports a nice pattern for CLI endpoints

    TODO: start handling level configuration from logging yaml config

    :param func:
    :param log_format:
    :return:
    """
    @wraps(func)
    def wrapper(*args, **kwargs):
        parsed_args = func(*args, **kwargs)
        level = logging.INFO

        if hasattr(parsed_args, "verbose") and parsed_args.verbose:
            level = logging.DEBUG

        logging.basicConfig(
            level=level,
            format=log_format,
            datefmt="%d-%m-%y %T",
        )

        logging.getLogger("cdsapi").setLevel(logging.WARNING)
        logging.getLogger("matplotlib").setLevel(logging.WARNING)
        logging.getLogger("matplotlib.pyplot").setLevel(logging.WARNING)
        logging.getLogger("requests").setLevel(logging.WARNING)
        logging.getLogger("tensorflow").setLevel(logging.WARNING)
        logging.getLogger("urllib3").setLevel(logging.WARNING)
        return parsed_args
    return wrapper





"""
Outlined in this section we will discuss the usage of the automated plotting functionallity
of the pyRoutePlanner. In this series of class distributions we house our interactive web based plots as well 
static plots ready for publication usage.
representation of input data. In each CellBox we determine the mean and variance of 
the information goverining our nemerical world, this includes and is not limited to:
Ocean Currents, Sea Ice Concentration, Bathemetric depth, whether on land.

Example:
    An example of running this code can be executed by running the following in a ipython/Jupyter Notebook::

        from RoutePlanner import CellBox
        ....

Additional information on constructing document strings using the Google DocString method can be found at
https://sphinxcontrib-napoleon.readthedocs.io/en/latest/example_google.html

Attributes:
    Some of the key attributes that the CellBox comprises are ...

Todo:
    * Adding the addition of ...


"""





import copy
import json
import pandas as pd
import matplotlib
import numpy as np
import sys
from branca.colormap import linear
import folium
from pyproj import transform
from shapely import wkt
import geopandas as gpd
from folium import plugins
from shapely.geometry import Polygon

sys.path.insert(0, 'folium')
sys.path.insert(0, 'branca')

import branca
import folium

# Adapted from: https://nbviewer.org/gist/BibMartin/f153aa957ddc5fadc64929abdee9ff2e
from branca.element import MacroElement
from jinja2 import Template
from matplotlib.collections import LineCollection
from mpl_toolkits.axes_grid1.inset_locator import inset_axes


class BindColormap(MacroElement):
    """Binds a colormap to a given layer.

    Parameters
    ----------
    colormap : branca.colormap.ColorMap
        The colormap to bind.
    """
    def __init__(self, layer, colormap):
        super(BindColormap, self).__init__()
        self.layer = layer
        self.colormap = colormap
        self._template = Template(u"""
        {% macro script(this, kwargs) %}
            {{this.colormap.get_name()}}.svg[0][0].style.display = 'block';
            {{this._parent.get_name()}}.on('layeradd', function (eventLayer) {
                if (eventLayer.layer == {{this.layer.get_name()}}) {
                    {{this.colormap.get_name()}}.svg[0][0].style.display = 'block';
                }});
            {{this._parent.get_name()}}.on('layerremove', function (eventLayer) {
                if (eventLayer.layer == {{this.layer.get_name()}}) {
                    {{this.colormap.get_name()}}.svg[0][0].style.display = 'none';
                }});
        {% endmacro %}
        """)  # noqa



def deep_search(needles, haystack):
    found = {}
    if type(needles) != type([]):
        needles = [needles]

    if type(haystack) == type(dict()):
        for needle in needles:
            if needle in haystack.keys():
                found[needle] = haystack[needle]
            elif len(haystack.keys()) > 0:
                for key in haystack.keys():
                    result = deep_search(needle, haystack[key])
                    if result:
                        for k, v in result.items():
                            found[k] = v
    elif type(haystack) == type([]):
        for node in haystack:
            result = deep_search(needles, node)
            if result:
                for k, v in result.items():
                    found[k] = v
    return found



from shapely.geometry.polygon import Polygon
import cartopy.crs as ccrs
import matplotlib.pyplot as plt
from shapely import wkt
import cartopy.feature as cfeature
import pandas as pd
import cartopy.io.img_tiles as cimgt
import json


class StaticMap:
    def __init__(self, json):
    
        self.json = json
        self.config  = json['config']
        self.basemap = self.config["Static_Map"]['Basemap_Info']
        self.layers  = self.config['Static_Map']['Layers']
        cellboxes    = json['cellboxes']
        


        #Initialising the basemap
        if 'CRS' in self.basemap:
            if self.basemap['CRS'] == 'Mercartor':
                self.ccrs = ccrs.Mercator()
            if self.basemap['CRS'] == 'Orthographic':
                self.ccrs = ccrs.Orthographic(central_longitude=self.basemap['central_longitude'],central_latitude=self.basemap['central_latitude'])
        else:
           self.ccrs = ccrs.Mercator()
        self._basemap()


        # Overlaying the layers
        for idx, layer in enumerate(self.layers):
            self.zorder = idx+1
            if not layer['Show']:
                continue
            #try:
            if layer['Type'] == 'Maps':
                self._maps(layer, cellboxes)
            if layer['Type'] == 'Paths':
                self._paths(layer)
            if layer['Type'] == 'Points':
                self._points(layer)

        if 'Output_Filename' in self.config['Static_Map']:
            plt.savefig(self.config['Static_Map']['Output_Filename'])


    def _basemap(self):
        self.fig = plt.figure(figsize=(15,15))
        matplotlib.rcParams.update({'font.size': 16})

        self.ax = plt.axes(projection=self.ccrs)
        self.ax.set_extent([self.config['Mesh_info']['Region']['longMin']+1e-6,self.config['Mesh_info']['Region']['longMax']-1e-6,self.config['Mesh_info']['Region']['latMin'],self.config['Mesh_info']['Region']['latMax']], crs=ccrs.PlateCarree())
        self.ax.add_image(cimgt.GoogleTiles(), 3)
        self.ax.coastlines(resolution='50m')
        self.ax.gridlines(draw_labels=True, dms=True, x_inline=False, y_inline=False,linewidth=0.5,linestyle='--')
        self.ax.add_feature(cfeature.BORDERS)
        plt.title(r'Route Planner | {}'.format(self.basemap['Title']),fontsize=14,loc='left',color='blue')

    def _points(self,info):

        if 'filename' in info:
            dataframe_points = pd.read_csv(info['filename'])
        elif 'object' in info:
            dataframe_points = pd.DataFrame(deep_search([info['object']],self.json )[info['object']])
        else:
            print('Please define either a filename or object in the json')

        if ('Name' in info) and info['Name']:
            dataframe_points[info['Name']][dataframe_points[info['Name']] > info['Fill_trim_max']] = np.nan
            dataframe_points[info['Name']][dataframe_points[info['Name']] <= info['Fill_trim_min']] = np.nan
            dataframe_points = dataframe_points.dropna()
            if ('Cmap' in info) and info['Cmap']:
                cmap = linear._colormaps[info['Cmap']].scale(dataframe_points[info['Name']].min(), dataframe_points[info['Name']].max())
                colour = [cmap.rgba_floats_tuple(ii) for ii in np.array(dataframe_points[info['Name']])]
            else:
                colour = info['Color']
            self.ax.scatter(dataframe_points['Long'],dataframe_points['Lat'],info["Size"],c=colour,marker='o',transform=ccrs.PlateCarree(),zorder=self.zorder)
        else:
            self.ax.scatter(dataframe_points['Long'],dataframe_points['Lat'],info["Size"],marker='o',transform=ccrs.PlateCarree(),color=info['Color'],zorder=self.zorder)


    def _paths(self,info):
        '''
            Plotting a paths type object
        '''

        # Loading the path information

        if 'filename' in info:
            with open(info['filename'], 'r') as f:
                geojson = json.load(f)
        elif 'object' in info:
            geojson = deep_search([info['object']],self.json)[info['object']]
        else:
            print('Please define either a filename or object in the json')
        paths = geojson['features']

        # Determining max values of all paths
        if info['Colorline']:
            max_val = 0
            min_val = 0
            for path in copy.deepcopy(paths):
                if np.array(path['properties'][info['Data_Name']]).max() > max_val:
                    max_val = np.array(path['properties'][info['Data_Name']]).max()
                if np.array(path['properties'][info['Data_Name']]).min() < min_val:
                    min_val = np.array(path['properties'][info['Data_Name']]).min()

        # Determining max travel-times of all paths
        for path in paths:
            points   = np.array(path['geometry']['coordinates'])
            if 'Data_Name' in info:
                data_val = np.array(path['properties'][info['Data_Name']])
            else:
                data_val = np.array(len(points))


            if info['Colorline']:
                # Add ColorLines
                x = self.ccrs.transform_points(x=points[:,0], y=points[:,1],
                                                src_crs=ccrs.PlateCarree())
                xcs = np.array([x[:,0],x[:,1]]).T.reshape(-1,1,2) 
                segments = np.concatenate([xcs[:-1], xcs[1:]], axis=1)   
                lc = LineCollection(segments, cmap=info['Cmap'], linewidth=info['Line_Width'],norm=plt.Normalize(vmin=min_val, vmax=max_val),zorder=self.zorder) 
                lc.set_array(path['properties']['traveltime'])                                           
                self.ax.add_collection(lc) 
                
            
            else:
                self.ax.plot(points[:,0],points[:,1],transform=ccrs.PlateCarree(),linewidth=info['Line_Width'],color=info['Color'],alpha=1.,zorder=self.zorder)

            if info['Path_Points']:
                self.ax.scatter(points[:,0],points[:,1],color=info['Color'],)
        
        if info['Colorline']:
            cbaxes = inset_axes(self.ax, '25%', '3%', loc =1)
            cbaxes.set_facecolor([1,1,1,0.7])
            cb=self.fig.colorbar(lc,cax=cbaxes,orientation='horizontal',label='Traveltime') #make colorbar



    def _maps(self,info, cellboxes):
            '''
                Plotting a map type object
            '''
            dataframe_pandas = pd.DataFrame(cellboxes)
            #dataframe_pandas = pd.read_csv(info['filename'])
            dataframe_pandas['geometry'] = dataframe_pandas['geometry'].apply(wkt.loads)
            dataframe_geo = gpd.GeoDataFrame(dataframe_pandas,crs='EPSG:4326', geometry='geometry')

            if dataframe_geo[info['Data_Name']].dtype == 'bool':
                for _,poly in dataframe_geo.iterrows():
                    if poly[info['Data_Name']]:
                        self.ax.add_geometries([poly['geometry']], crs=ccrs.PlateCarree(), edgecolor=info['Line_Color'], facecolor=info['Fill_Color'],alpha=info['Fill_Opacity'],lw=info['Line_Width'],zorder=self.zorder)


            if dataframe_geo[info['Data_Name']].dtype == float:
                if ('Fill_trim_min' not in info):
                    info['Fill_trim_min'] = dataframe_geo[info['Data_Name']].min()
                if ('Fill_trim_max' not in info):
                    info['Fill_trim_max'] = dataframe_geo[info['Data_Name']].max()

                dataframe_geo = dataframe_geo[(dataframe_geo[info['Data_Name']] >= info['Fill_trim_min']) &
                                            (dataframe_geo[info['Data_Name']] <= info['Fill_trim_max'])]

                if info['Fill_cmap_use']:
                    if 'Fill_cmap_opacity' not in info:
                        info['Fill_cmap_opacity'] = False
                    if info['Fill_cmap_opacity']:
                        for _,poly in dataframe_geo.iterrows():
                            self.ax.add_geometries([poly['geometry']], crs=ccrs.PlateCarree(), edgecolor=info['Line_Color'], facecolor=info['Fill_Color'],alpha=poly[info['Data_Name']]/info['Fill_cmap_opacity_scalar'],lw=info['Line_Width'],zorder=self.zorder)
                    else:
                        if 'Fill_cmap_min' in info:
                            cmin = info['Fill_cmap_min']
                        else:
                            cmin = dataframe_geo[info['Data_Name']].min()
                        if 'Fill_cmap_max' in info:
                            cmax = info['Fill_cmap_max']
                        else:
                            cmax = dataframe_geo[info['Data_Name']].max()
                            # Define a colormap version
                            cmap = linear._colormaps[info['Cmap']].scale(info['Fill_trim_min'], info['Fill_trim_max'])
                            for indx,row in dataframe_geo.iterrows():
                                data = row[info['Data_Name']]
                                if not np.isnan(data):
                                    colour = cmap.rgba_floats_tuple(data)
                                    self.ax.add_geometries([row['geometry']], facecolor = colour, crs = ccrs.PlateCarree(), alpha = info['Fill_Opacity'],zorder=self.zorder)
                else:
                    if (info['Fill_Opacity'] == 0.0):
                        for _,poly in dataframe_geo.iterrows():
                            x,y = poly['geometry'].exterior.coords.xy
                            self.ax.plot(np.array(x),np.array(y),color=info['Line_Color'],linewidth=info['Line_Width'],transform=ccrs.PlateCarree(),zorder=self.zorder)
    def show(self):
        plt.show()


class InteractiveMap:
    def __init__(self,config):
        self.config = config

        self.basemap = config["Interactive_Map"]['Basemap_Info']
        self.layers  = config['Interactive_Map']['Layers']


        # Initialising the basemap
        self._basemap()


        # Defining the layers to plot
        for layer in self.layers:
            try:
                if layer['Type'] == 'Paths':
                    self._paths(layer)
                if layer['Type'] == 'Maps':
                    self._maps(layer) 
                if layer['Type'] == 'Points':
                    self._points(layer)
                if layer['Type'] == 'Geotiff':
                    self._geotiff(layer)
                if layer['Type'] == 'MeshInfo':
                    self._meshInfo(layer)
            except:
                print('Issue Plotting Layer')
                    
        # Adding in the layer control
        self._layer_control()

        # Save
        if 'Output_Filename' in self.config['Interactive_Map']:
            self.map.save(self.config['Interactive_Map']['Output_Filename'])


    def _basemap(self):
        '''
            Defining the basemap structure
        '''
        title_html = '''
            <h1 style="color:#003b5c;font-size:16px">
            &ensp;<img src='https://i.ibb.co/JH2zknX/Small-Logo.png' alt="BAS-colour-eps" border="0" style="width:40px;height:40px;"> 
            <img src="https://i.ibb.co/XtZdzDt/BAS-colour-eps.png" alt="BAS-colour-eps" border="0" style="width:179px;height:40px;"> 
            &ensp;|&ensp; RoutePlanner &ensp;|&ensp; {}
            </h1>
            </body>
            '''.format(self.basemap['Title'])   
        self.map = folium.Map(location=self.basemap['Map_Centre'],zoom_start=self.basemap['Map_Zoom'],tiles=None)
        
        # 
        bsmap = folium.FeatureGroup(name='BaseMap')
        folium.TileLayer('https://server.arcgisonline.com/ArcGIS/rest/services/NatGeo_World_Map/MapServer/tile/{z}/{y}/{x}.png',attr="toner-bcg", name='Basemap').add_to(bsmap)
        bsmap.add_to(self.map)
        bsmap = folium.FeatureGroup(name='Dark BaseMap',show=False)
        folium.TileLayer(tiles="https://{s}.basemaps.cartocdn.com/rastertiles/dark_all/{z}/{x}/{y}.png",attr='&copy; <a href="https://www.openstreetmap.org/copyright">OpenStreetMap</a> contributors &copy; <a href="https://carto.com/attributions">CARTO</a>',name='darkmatter').add_to(bsmap)
        bsmap.add_to(self.map)



        self.map.get_root().html.add_child(folium.Element(title_html))
        
    def _paths(self,info):
        '''
            Plotting a paths type object
        '''

        #map,PathPoints=True,PathName='Transit Time', colorLine=True
        
        # Defining the feature groups to add
        pths        = folium.FeatureGroup(name='{}'.format(info['Name']),show=info['Show'])
        if info['Path_Points']:
            pths_points = folium.FeatureGroup(name='{} - Path Points'.format(info['Name']),show=info['Show'])
            
        # Loading the path information
        with open(info['filename'], 'r') as f:
            geojson = json.load(f)
            paths = geojson['features']

        # Determining max values of all paths
        if info['Colorline']:
            max_val = 0
            for path in copy.deepcopy(paths):
                if np.array(path['properties'][info['Data_Name']]).max() > max_val:
                    max_val = np.array(path['properties'][info['Data_Name']]).max()

        # Determining max travel-times of all paths
        for path in paths:
            points   = np.array(path['geometry']['coordinates'])
            if 'Data_Name' in info:
                data_val = np.array(path['properties'][info['Data_Name']])
            else:
                data_val = np.array(len(points))

            points[:,0] = points[:,0]
            points = points[:,::-1]

            if info['Colorline']:
                #folium.PolyLine(points,color="black", weight=info['Line_Width'], opacity=1).add_to(pths)
                colormap = linear._colormaps[info['Cmap']].scale(0,max_val)
                colormap.caption = '{} ({},Max Value={:.3f})'.format(info['Data_Name'],info['Data_Units'],max_val)
                folium.ColorLine(points,data_val,colormap=colormap,nb_steps=50, weight=info['Line_Width'], opacity=1).add_to(pths)



            else:
                folium.PolyLine(points,color=info['Color'], weight=info['Line_Width'], opacity=1).add_to(pths)


            if info['Path_Points']:
                for idx in range(len(points)):
                    loc = [points[idx,0],points[idx,1]]
                    folium.Marker(
                        location=loc,
                        icon=folium.plugins.BeautifyIcon(icon='circle',
                                                    border_color='transparent',
                                                    background_color='transparent',
                                                    border_width=1,
                                                    text_color=info['Color'],
                                                    inner_icon_style='margin:0px;font-size:0.8em')
                    ).add_to(pths_points)
        
        if info['Path_Points']:
            pths_points.add_to(self.map)
        if info['Colorline']:
            self.map.add_child(pths)
            self.map.add_child(colormap)
            self.map.add_child(BindColormap(pths,colormap))
        else:
            pths.add_to(self.map)


    def _points(self,info):

        wpts = folium.FeatureGroup(name='{}'.format(info['Name']),show=info['Show'])
        wpts_name = folium.FeatureGroup(name='{} - Names'.format(info['Name']),show=info['Show'])

        dataframe_points = pd.read_csv(info['filename'])


        for id,wpt in dataframe_points.iterrows():
            loc = [wpt['Lat'], wpt['Long']]
            folium.Marker(
                location=loc,
                icon=plugins.BeautifyIcon(icon='circle',
                                                border_color='transparent',
                                                background_color='transparent',
                                                border_width=info['Border_Width'],
                                                text_color=info['Color'],
                                                inner_icon_style='margin:0px;font-size:0.8em'),
                popup="<b>{} [{:4f},{:4f}]</b>".format(wpt['Name'],loc[0],loc[1]),
            ).add_to(wpts)    

            folium.Marker(
                        location=loc,
                            icon=folium.features.DivIcon(
                                icon_size=(250,36),
                                icon_anchor=(0,0),
                                html='<div style="font-size: {}pt">{}</div>'.format(info['Font_Size'],wpt['Name']),
                                ),
            ).add_to(wpts_name)


        wpts.add_to(self.map)
        wpts_name.add_to(self.map)


    def _maps(self,info):
        '''
            Plotting a map type object
        '''

        dataframe_pandas = pd.read_csv(info['filename'])
        dataframe_pandas['geometry'] = dataframe_pandas['geometry'].apply(wkt.loads)
        dataframe_geo = gpd.GeoDataFrame(dataframe_pandas,crs='EPSG:4326', geometry='geometry')


        feature_info = folium.FeatureGroup(name='{}'.format(info['Name']),show=info['Show'])

        if dataframe_geo[info['Data_Name']].dtype == 'bool':
            folium.GeoJson(
                dataframe_geo[dataframe_geo[info['Data_Name']]==True],
                style_function=lambda x: {
                    'fillColor': info['Fill_Color'],
                    'color': info['Line_Color'],
                    'weight': info['Line_Width'],
                    'fillOpacity': info['Fill_Opacity']
                    }
            ).add_to(feature_info)

            feature_info.add_to(self.map)

        if dataframe_geo[info['Data_Name']].dtype == float:

            if ('Fill_trim_min' not in info):
                info['Fill_trim_min'] = dataframe_geo[info['Data_Name']].min()
            if ('Fill_trim_max' not in info):
                info['Fill_trim_max'] = dataframe_geo[info['Data_Name']].max()            

            dataframe_geo = dataframe_geo[(dataframe_geo[info['Data_Name']] >= info['Fill_trim_min']) &
                                        (dataframe_geo[info['Data_Name']] <= info['Fill_trim_max'])]

            if info['Fill_cmap_use']:

                if 'Fill_cmap_opacity' not in info:
                    info['Fill_cmap_opacity'] = False

                if info['Fill_cmap_opacity']:
                    folium.GeoJson(
                        dataframe_geo,
                        style_function=lambda x: {
                                'fillColor': info['Fill_Color'],
                                'color': info['Line_Color'],
                                'weight': info['Line_Width'],
                                'fillOpacity': x['properties'][info['Data_Name']]/info['Fill_cmap_opacity_scalar']
                            }
                    ).add_to(feature_info)
                    self.map.add_child(feature_info)

                else:
                    if 'Fill_cmap_min' in info:
                        cmin = info['Fill_cmap_min']
                    else:
                        cmin = dataframe_geo[info['Data_Name']].min()
                    if 'Fill_cmap_max' in info:
                        cmax = info['Fill_cmap_max']
                    else:
                        cmax = dataframe_geo[info['Data_Name']].max()

                    colormap = linear._colormaps[info['Cmap']].scale(cmin,cmax)
                    colormap.caption = '{} ({})'.format(info['Data_Name'],info['Data_Units'])
                    folium.GeoJson(
                        dataframe_geo,
                        style_function=lambda x: {
                                'fillColor': colormap(x['properties'][info['Data_Name']]),
                                'color': info['Line_Color'],
                                'weight': info['Line_Width'],
                                'fillOpacity': info['Fill_Opacity']
                            }
                    ).add_to(feature_info)
                    self.map.add_child(feature_info)
                    self.map.add_child(colormap)
                    self.map.add_child(BindColormap(feature_info,colormap))
            else:
                folium.GeoJson(
                    dataframe_geo,
                    style_function=lambda x: {
                        'fillColor': info['Fill_Color'],
                        'color': info['Line_Color'],
                        'weight': info['Line_Width'],
                        'fillOpacity': info['Fill_Opacity']
                        }
                ).add_to(feature_info)

                self.map.add_child(feature_info)

    #   {
    #     "Type": "MeshInfo",
    #     "Show":false,
    #     "Name": "Mesh Info",
    #     "filename": "./cellgrid_dataframe.csv",
    #     "Line_Width": 0.3,
    #     "Line_Color": "gray",
    #     "Fill_Color": "white",
    #     "Fill_Opacity": 0.0,
    #     "Data_Names":["cell_info"]
    #   },

    def _meshInfo(self,info):
        dataframe_pandas = pd.read_csv(info['filename'])
        dataframe_pandas['geometry'] = dataframe_pandas['geometry'].apply(wkt.loads)
        dataframe_geo = gpd.GeoDataFrame(dataframe_pandas,crs='EPSG:4326', geometry='geometry')
        dataframe_geo = dataframe_geo[['geometry']+list(info['Data_Names'])]

        feature_info = folium.FeatureGroup(name='{}'.format(info['Name']),show=info['Show'])
        folium.GeoJson(dataframe_geo,
            style_function= lambda x: {
                    'fillColor': info['Fill_Color'],
                    'color': info['Line_Color'],
                    'weight': info['Line_Width'],
                    'fillOpacity': info['Fill_Opacity']
                },
            tooltip=folium.GeoJsonTooltip(
                fields=list(dataframe_geo.columns.intersection(info['Data_Names'])),
                aliases=info['Label_Names'],
                localize=True
            )
        ).add_to(feature_info) 
        feature_info.add_to(self.map)
    def _layer_control(self):
        folium.LayerControl(collapsed=True).add_to(self.map)

    def show(self):
        return self.map


    def _geotiff(self,info):
        from rasterio.windows import Window
        import rasterio as rs
        from pyproj import Proj
        from math import floor, ceil
        from rasterio.transform import guard_transform
        import numpy as np

        def windowFunc(lon, lat, dataset):
            p = Proj(dataset.crs)
            t = dataset.transform
            xmin, ymin = p(lon[0], lat[0])
            xmax, ymax = p(lon[1], lat[1])
            col_min, row_min = ~t * (xmin, ymin)
            col_max, row_max = ~t * (xmax, ymax)


            window = Window.from_slices(rows=(floor(row_max), ceil(row_min)),cols=(floor(col_min), ceil(col_max)))
            transform = guard_transform(dataset.transform)
            return window,rs.windows.bounds(window, transform)

        dataset = rs.open(info['filename'])
        resamplingFactor = info['Resampling_Factor']
        window=None
        bnds = [[dataset.bounds[1], dataset.bounds[0]], [dataset.bounds[3], dataset.bounds[2]]]
        trying =True
        indx = 1
        while trying:
            try:
                img = folium.raster_layers.ImageOverlay(
                    name="{} - Band {}".format(info['Name'],indx),
                    image=dataset.read(indx,window=window)[::resamplingFactor,::resamplingFactor],
                    bounds=bnds,
                    opacity=info['Opacity'],
                    mercator_project=True,
                    pixelated=False,
                    show = info['Show']
                )
                img.add_to(self.map)
            except:
                trying=False
            indx+=1


from setuptools import setup, find_packages

import polar_route


def get_content(filename):
    with open(filename, "r") as fh:
        return fh.read()


requirements = get_content("requirements.txt")

setup(
    name=polar_route.__name__,
    version=polar_route.__version__,
    description=polar_route.__description__,
    license=polar_route.__license__,
    long_description=get_content("README.md"),
    long_description_content_type="text/markdown",
    author=polar_route.__author__,
    author_email=polar_route.__email__,
    maintainer=polar_route.__author__,
    maintainer_email=polar_route.__email__,
    url="https://www.github.com/antarctica",
    project_urls={
    },
    classifiers=[el.lstrip() for el in """
        Development Status :: 3 - Alpha
        Intended Audience :: Science/Research
        Intended Audience :: System Administrators
        License :: OSI Approved :: MIT License
        Natural Language :: English
        Operating System :: OS Independent
        Programming Language :: Python
        Programming Language :: Python :: 3
        Programming Language :: Python :: 3.7
        Topic :: Scientific/Engineering
    """.split('\n')],
    entry_points={
        'console_scripts': [
            "create_mesh=polar_route.cli:create_mesh_cli",
            "add_vehicle=polar_route.cli:add_vehicle_cli",
            "optimise_routes=polar_route.cli:optimise_routes_cli"],
    },
    keywords=[],
    packages=find_packages(),
    install_requires=requirements,
    tests_require=["pytest"],
    extras_require={
        "tests": get_content("tests/requirements.txt"),
    },
    zip_safe=False,
    include_package_data=True)
